#########################################################################################
# PPP共通仕様協議会が定めるNGSI V2の規格に合致するJSON文をcsvに変換します
# Input   inputfile: 入力となるJSON文。UTF-8。
#         paramfile JSONからcsvに変換する際のattributeと項目との対応を定義するファイル。UTF-8
#                   形式はkey-Valureペアのであり、{<key-Valueペア> <key-Valueペア>,<key-Valueペア>
# 　　　　　　　　　　の形式。各key-Valueペアは、<Atribute指定>:<csv項目指定>の形式。
#                   <Attribute指定>は、Attribute nameまたは、Attribute nameとSub-attribute nameを
#                   カンマほ挟んで繋げた形式。尚、IdentificationGroupとGeoJSONは、前記の形式に加え、
#                   特定の内容を取り出す機能をサポートする。サポートする機能は以下の通り
#                   IdentificationGroup
#                      <Attribute指定>/IdentificationType : Identificationtypeに対応するIDを取り出す 
#                   GeoJSON
#                      <Attribute指定>/Point   : 座標点を取り出す
#                      <Attribute指定>/Polygon : ポリゴンを取り出す
#                   key-valueペアのValueは生成するcsvの項目名を指定する
# Output outputfile:出力するcsvファイル。UTF-8。paramfileに指定した項目順に項目は並ぶ。対応するAttributeが
#                   が無い場合は空文を出力する。
# オプション inputfile: inputfileのファイル名。既定値はLandTest.json
#         inputdir:  inputfileが存在するティレクトリのバス。既定値は"c:\test\"
#         outputfile: outputfileのファイル名。既定値は"Land-csv.csv"
#         outputdir: outputfileが存在するティレクトリのバス。既定値は"c:\test\"
#         paramfile: : paramfileのファイル名。既定値は"="Land-jtoctable.json"
#         paramdir: paramfileが存在するティレクトリのバス。既定値は"c:\test\"
#         outmode: 既存のoutputfileを置き換えるか、後ろに行を追加するかの指定
#                  outmode=create: 置き換える。最初の行には項目名が出力される(既定値)
#                  outmode=append: 追加する。最初の行に項目名を出力しない
#         jsontype: inputfileのJSON文の形式。JSONとしては以下の3種類をサポートする
#                      jsontype=entity : 波括弧に囲まれるひとつのEntity
#                      jsontype=array : 角かっこに囲まれるEntityの列  
#                      jsontype=batch : op/updateに対し指定する、一括登録のためのJSON文(既定値)

import json
import csv
import os
import sys
from my_package.NGSIv2 import GET_entities, imschema, entity, trace_log, error_log
from my_package.CoppellTechnologies_common import CT_csv, CT_json

#########################################################################################
# 各変数の初期化
def initialize():
    run = {
        'param': {
            'dir': 'c:/Tools/data/',
            'inputdir': '',
            'inputfile': 'Landjson.json',
            'outputdir': '',
            'outputfile': 'Landcsv2.csv',
            'outputtype': '',
            'outputmode': 'extend',
            'outmode': 'create',
            'schemadir': '',
            'schemafile': 'Land.imschema.json',
            'Organizationdir': '',
            'Organizationfile': '',
            'Departmentdir': '',
            'Departmentfile': 'Departmentjson.json',
            'Landdir': '',
            'Landfile': '',
            'Buildingdir': '',
            'Buildingfile': '',
            'Facilitydir': '',
            'Facilityfile': '',
            'FacilityStatisticsdir': '',
            'FacilityStatisticsfile': '',
            'BuildingComponentdir': '',
            'BuildingComponentfile': '',
            'Devicedir': '',
            'Devicefile': '',
            'DeviceModeldir': '',
            'DeviceModelfile': '',
            'Complaintdir': '',
            'Complaintfile': '',
            'Reportdir': '',
            'Reportfile': '',
            'addressdir': 'c:/Tools/data/',
            'addressfile': 'mt_city_all.csv',
            'logdir': '',
            'logfile': 'Landjtoclog.txt',
            'jsontype': 'array',
            'format': 'normalized',
            'truevalue': ['True', 'true', '真', '有'],
            'falsevalue': ['False', 'false', '偽', '無']
        },
        'handles': {},
        'errorflag': False,
        'onerror': 'continue',
        'types': ['Text','Number','Integer','Boolean','Date', 'DateTime','URL','Time','Relationship'],
        'entitynames': ['Organization','Department','Land','Building',
            'Facility','FacilityStatistics','BuildingComponent','Device',
            'DeviceModel','Complaint','Report'],
        'errorobj': {},
        'errorinfo': {'attr': ''},
        'warning': False,
        'type_tbl': {'str':'string', 'int':'integer','float':'number', 'bool':'boolean' }
    }
    param = run['param']
    if len(sys.argv) > 1:
        for arg in sys.argv[1:]:
            p = arg.split('=')
            if len(p) != 2:
                input(f'## error ## 不正なパラメータです: {arg}')
                exit()
            if p[0] in param.keys():
                param[p[0]] = p[1]
            else:
                input( f'## error ## 未知のパラメータです: {p[0]}')
                exit()
    if (param['inputdir'] == ''):
        param['inputdir'] = param['dir']
    if (param['outputdir'] == ''):
        param['outputdir'] = param['dir']
    if (param['schemadir'] == ''):
        param['schemadir'] = param['dir']
    if (param['outputdir'] == ''):
        param['outputdir'] = param['dir']
    if (param['addressdir'] == ''):
        param['addressdir'] = param['dir']
    for a1 in run['entitynames']:
        if ((a1 + 'dir') in param and param[a1 + 'dir'] == ''):
            param[a1 + 'dir'] = param['dir']
    if (param['logdir'] == ''):
        param['logdir'] = param['dir']
    return run

## ここからメインプログラムです
## まず、入出力ファイルの格納位置とファイル名を指定します。環境に応じて書き換えるか、何等かの方法
## で外部から渡してもらう様に変更する必要があります

## 変数を初期化します
if __name__ == '__main__':
    run = initialize()
    param = run['param']
    handles =run['handles']
    handles['error'] = err = error_log(run)
    handles['trace_log'] = trace_log(run)
    handles['entity'] = entity(run)
        
    ## 必須パラメータをチェックします
    if (param['inputfile'] == ''):
        err.set_errorobj('01', f'## error ## inputfile parameter missing')
    if (param['outputfile'] == ''):
        err.set_errorobj('02', f'## error ## outputfile parameter missing')
    if (param['schemafile'] == ''):
        err.set_errorobj('03', f'## error ## schemafile parameter missing')
    
    ## JSONスキーマと変換表を辞書として読み込みます
    handles['schema'] = imschema(run)
    
    ## 入力のJSON文を読み込みます
    filename = param['inputfile']
    directory = param['inputdir']
    try:
        entities = CT_json.load(directory, filename)
    except FileNotFoundError as e:
        err.set_errorobj('04', f'## error ## inputfileが見つからない. filename: {filename}, directory: {directory}')
    except UnicodeDecodeError as e:
        err.set_errorobj('05', f'## error ## inputfileのコードがUTF-8ではない. filename: {filename}, directory: {directory}')
    except csv.Error as e:
        err.set_errorobj('06', f'## error ## inputfileのcsvにエラーが見つかりました. filename: {filename}, directory: {directory}')
    except Exception as e:
        err.set_errorobj('07', f'## error ## inputfileのcsv読み込み時にエラーが発生しました. filename: {filename}, directory: {directory}, {str(e)}')
    if isinstance(entities,list):          # list形式にentityが並んでいる形式
        pass
    elif isinstance(entities,dict):        # entity単独かバッチ登録可能な形式
        if entities.get('entities'):       # バッチ形式
            entities = entities['entities']
        else:                              # 単独entity
            entities = [entities]
    else:
        err.set_errorobj('11', f'## error ## jsonファイルの形式が正しくない. filename: {filename}, directory: {directory}.')
    if not len(entities):
        err.set_errorobj('12', f'## error ## jsonファイルから読み込んだentityがゼロ件. filename: {filename}, directory: {directory}.')
    if not entities[0].get('id') or not entities[0].get('type'):
        err.set_errorobj('13', f'## error ## jsonファイルから読み込んだentityにidまたはtypeが無い. filename: {filename}, directory: {directory}.')
    handles['trace_log'].set_log(0,'08',f'input: directory: {directory}, inputfile: {filename}')
    handles['trace_log'].set_log(0,'09',f'output: outputdir:  {directory}, inputfile: {filename}')
    
    ## 各Entityのファイルを辞書のリストとして読み込みます
    try:
        for entityname in run['entitynames']:
            directory = entityname+'dir'
            filename = entityname+'file'
            if ((filename) in param and param[filename] != ''):
                handles[entityname] = GET_entities(run,entityname)
                handles[entityname].load(param[entityname + 'dir'],param[entityname + 'file'])
                handles['trace_log'].set_log(0,'08',f'referenced entities: directory: {directory}, {entityname}file: {filename}')
                if not isinstance(handles[entityname].entities,list):
                    err.set_errorobj('09', f'## error ## Entitiesの形式がリスト構造ではありません. Entity type: {entityname}')
                if len(handles[entityname].entities) == 0:
                    err.set_errorobj('10', f'## error ## Entitiesの件数がゼロ件です. Entity type: {entityname}')
                if handles[entityname].entities[0].get('id') == None:
                    err.set_errorobj('11', f'## error ## Entityの:形式が不正です. Entity type: {entityname}')
    except FileNotFoundError as e:
        handles['error'].set_errorobj('12', f'## error ## Entities load  error : {e}\n')
    
    ## 出力する2次元のリストの一行目(ヘッダ)を作成します
    out_list = []
    if (param['outmode']=='create'):
        out_list.append(handles['schema'].get_header(run['scm']))
    
    ## 出力する2次元のリストの二行目以降を作成します
    for in_entity in entities:
        handles['trace_log'].set_log(0,'13',f'Entity# {len(out_list)}')
        out_list.append(run['handles']['schema'].get_row(run['scm'], in_entity))
    
    ## 出来上がった二次元のリストをcsvファイルに出力します
    if (param['outmode'] == 'create'):
        outmode='w'
    elif (param['outmode'] =='append'):
        outmode='a'
    else:
        err.set_errorobj('13', f'パラメータのoutmodeが正しくありません')
    filename = param['outputfile']
    directory = param['outputdir']
    try:
        CT_csv.save(directory, filename, out_list, mode='w')
    except FileNotFoundError as e:
        err.set_errorobj('14', f'## error ## {directory} ディレクトリが見つかりません: {e}')
    except PermissionError as e:
        err.set_errorobj('15', f'{filename} ファイルがオープンできません.\nファイル名を変更するか、このファイルを使っているプログラムを終了してください')
    
    if (run['errorflag']==True):
        handles['trace_log'].set_log(0,'16',f'CSV to JSON translation faileded')
        print('Error occurd. Please, check message, parameters and files\n')
        input(run['errorobj'])
    if run['warning'] == True:
        handles['trace_log'].__del__()
        input(f'jsonからcsvへの変換中にワーニングが発生しました。ログを確認してください\n')
    else:
        handles['trace_log'].set_log(0,'17',f'JSON translated to CSV successfully. Entity number: '+str(len(out_list)-1))
        handles['trace_log'].__del__()
        print(f'json translated to csv successfully. output CSV line number: {str(len(out_list))}\n')