#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# checked with 3.10
# Created by H. Inoue on 2024-02-10 
"""
Coppell Technologiesの各種ツールから利用する共通関数群です
本クラス群は汎用的に利用される事を想定しています。
基本的に例外は処理していないため、呼び出し側でハンドリングが必要です
"""
import os
import json
import csv
import re
## Coppell Technologiesが提供するツール群の共通クラスです

class CT_json:
    ## json関連共通クラスです
    
    ## ファイル内のJSON文をdictに展開します。返却値は展開したdictです
    @classmethod
    def load (cls, directory, filename):
        if directory != '':
            os.chdir(directory)
        with open(filename, 'r', encoding='utf-8') as f:
            return json.loads(f.read())

class CT_csv:
    ## csv関連共通クラスです
    
    ## ファイル内のcsvを二次元のlistに展開します。返却値は展開したlistです
    @classmethod
    def load(cls, directory, filename):
        os.chdir(directory)
        with open(filename, 'r', encoding='utf-8') as f:
            reader = csv.reader(f)
            tbl = [row for row in reader]
        # Excelでcsvを作成した場合に付加されるBOMです
        if tbl and tbl[0][0].startswith('\ufeff'):
            tbl[0][0] = tbl[0][0][1:]
        return tbl
    
    @classmethod
    def save(cls, directory, filename, data, mode='w'):
        if directory != '':
            os.chdir(directory)
        with open(filename, mode, newline='', encoding='utf-8') as f:
            writer = csv.writer(f)
            writer.writerows(data)

class CT_text:
    ## text関連共通クラスです
    
    ## ファイル内のtextを出力します。モードの既定値はwriteです
    @classmethod
    def save(cls, directory, filename, text, mode='w'):
    ## テキストをファイルとして出力します
        if directory != '':
            os.chdir(directory)
        with open(filename, mode, encoding='utf-8') as f:
            return f.write(text)                          # nは使いません。nは書き込んだ文字数です
    
    ## 文字列が数値に変換可能が確認します
    @classmethod
    def isnumeric(cls, text)->bool:
    ## テキストを数値に変換可能か調べます
        return re.match(r'^\d+(\.\d+)?$', text) is not None
    
    ## 文字列が数値に変換可能が確認します
    @classmethod
    def float(cls, text)->float:
    ## テキストを数値に変換して返却します
        if '.' in text:
            return float(text)
        else:
            return int(text)

