#########################################################################################
# PPP共通仕様協議会が公開しているNGSI V2の規格に準拠するJSON文をマージして、指定されたタイプ
# のEntityを出力します。
# Input   inputfile1: 入力となるJSONファイル。UTF-8。
#         inputfile2: 入力となるJSONファイル。UTF-8。
#          ・・・・・・
#         inputfile8: 入力となるJSONファイル。UTF-8。
# Output  outputfile:出力するJSONファイル。UTF-8。
# オプション
#         type: 出力するEntityのtype

import json
import os
import sys
from my_package.NGSIv2 import trace_log, error_log
from my_package.CoppellTechnologies_common import CT_json

#########################################################################################
# 各変数の初期化
def initialize():
    run = {
        'param': {
            'dir': 'c:/Tools/data/',
            'inputdir1': '',
            'inputfile1': '',
            'inputdir2': '',
            'inputfile2': '',
            'inputdir3': '',
            'inputfile3': '',
            'inputdir4': '',
            'inputfile4': '',
            'inputdir5': '',
            'inputfile5': '',
            'inputdir6': '',
            'inputfile6': '',
            'inputdir7': '',
            'inputfile7': '',
            'inputdir8': '',
            'inputfile8': '',
            'outputdir': '',
            'outputfile': '',
            'outputmode': 'extend',
            'logdir': '',
            'logfile': '',
            'jsontype': 'array',
            'format': 'normalized',
            'type': ''
        },
        'handles': {},
        'errorflag': False,
        'onerror': 'continue',
        'types': ['Text','Number','Integer','Boolean','Date', 'DateTime','URL','Time','Relationship'],
        'entitynames': ['Organization','Department','Land','Building',
            'Facility','FacilityStatistics','BuildingComponent','Device',
            'DeviceModel','Complaint','Report'],
        'errorobj': {},
        'errorinfo': {'attr': ''},
        'warning': False,
        'type_tbl': {'str':'string', 'int':'integer','float':'number', 'bool':'boolean' }
    }
    param = run['param']
    if len(sys.argv) > 1:
        for arg in sys.argv[1:]:
            p = arg.split('=')
            if len(p) != 2:
                input(f'## error ## 不正なパラメータです: {arg}')
                exit()
            if p[0] in param.keys():
                param[p[0]] = p[1]
            else:
                input( f'## error ## 未知のパラメータです: {p[0]}')
                exit()
    for i in range(8):
        if (param['inputfile'+str(i+1)] != ''):
            param['inputdir'+str(i+1)] = param['dir']
    if (param['outputdir'] == ''):
        param['outputdir'] = param['dir']
    if (param['logdir'] == ''):
        param['logdir'] = param['dir']
    return run

## ここからメインプログラムです
## まず、入出力ファイルの格納位置とファイル名を指定します。環境に応じて書き換えるか、何等かの方法
## で外部から渡してもらう様に変更する必要があります

## 変数を初期化します
if __name__ == '__main__':
    run = initialize()
    param = run['param']
    handles =run['handles']
    handles['error'] = err = error_log(run)
    handles['trace_log'] = trace_log(run)
        
    ## 必須パラメータをチェックします
    if (param['inputfile1'] == ''):
        err.set_errorobj('01', f'## error ## inputfile1 parameter missing')
    if (param['outputfile'] == ''):
        err.set_errorobj('03', f'## error ## outputfile parameter missing')
    if (param['type'] == ''):
        err.set_errorobj('04', f'## error ## type parameter missing')
    directory = param['inputdir1']
    filename = param['inputfile1']
    print (f'dir: {directory}, file: {filename}')
    try:
        in_entities = CT_json.load(directory, filename)
    except FileNotFoundError as e:
        err.set_errorobj('05', f'## error ## {filename} ファイルが見つからない. directory: {directory}')
    except UnicodeDecodeError as e:
        err.set_errorobj('06', f'## error ## {filename} jsonファイルのコードがUTF-8ではない. directory: {directory}')
    except json.Error as e:
        err.set_errorobj('07', f'## error ## jsonファイルにエラーが見つかりました. filename: {filename}, directory: {directory}')
    except Exception as e:
        err.set_errorobj('08', f'## error ## jsonファイルの読み込み時にエラーが発生しました. filename: {filename}, directory: {directory}.')
    # entitiesの形式を確認して、リストの先頭を探します
    if isinstance(in_entities,list):          # list形式にentityが並んでいる形式
        entities1 = in_entities
    elif isinstance(in_entities,dict):        # entity単独かバッチ登録可能な形式
        if in_entities.get('entities'):       # バッチ形式
            entities1 = in_entities['entities']
        else:                              # 単独entity
            entities1 = [in_entities]
    else:
        err.set_errorobj('11', f'## error ## jsonファイルの形式が正しくない. filename: {filename}, directory: {directory}.')
    if not len(entities1):
        err.set_errorobj('12', f'## error ## jsonファイルから読み込んだentityがゼロ件. filename: {filename}, directory: {directory}.')
    if not entities1[0].get('id') or not entities1[0].get('type'):
        err.set_errorobj('13', f'## error ## jsonファイルから読み込んだentityにidまたはtypeが無い. filename: {filename}, directory: {directory}.')
    handles['trace_log'].set_log(0,'01',f'input1: directory: {directory}, inputfile: {filename}')
    
    ## entity typeによるフィルタ
    entity_type = param['type']
    for i in range(len(entities1)-1, -1,-1):
        if entities1[i]['type'] != entity_type:
            del entities1[i]
            continue
        if i == 0:
            continue
        matched = False
        for j in range(i-1, -1, -1):
            add_ent = entities1[j]
            if entities1[j]['type'] != entity_type or entities1[j]['id'] != entities1[i]['id']:
                continue
            entities1[j].update(entities1[i])
            del entities1[i]
            matched = True
            break
        if matched:
            continue
    
    for i in range(1,8):
        if (param['inputfile'+str(i+1)] == ''):
            break
        directory = param['inputdir'+str(i+1)]
        filename = param['inputfile'+str(i+1)]
        try:
            add_entities = CT_json.load(directory, filename)
        except FileNotFoundError as e:
            err.set_errorobj('14', f'## error ## {filename} ファイルが見つからない. directory: {directory}')
        except UnicodeDecodeError as e:
            err.set_errorobj('15', f'## error ## {filename} jsonファイルのコードがUTF-8ではない. directory: {directory}')
        except json.Error as e:
            err.set_errorobj('16', f'## error ## jsonファイルにエラーが見つかりました. filename: {filename}, directory: {directory}')
        except Exception as e:
            err.set_errorobj('17', f'## error ## jsonファイルの読み込み時にエラーが発生しました. filename: {filename}, directory: {directory}.')
        # entitiesの形式を確認して、リストの先頭を探します
        if isinstance(add_entities,list):          # list形式にentityが並んでいる形式
            entities2 = add_entities
        elif isinstance(add_entities,dict):        # entity単独かバッチ登録可能な形式
            if in_entities.get('entities'):       # バッチ形式
                entities2 = add_entities['entities']
            else:                              # 単独entity
                entities2 = [in_entities]
        else:
            err.set_errorobj('11', f'## error ## jsonファイルの形式が正しくない. filename: {filename}, directory: {directory}.')
        if not len(entities2):
            err.set_errorobj('12', f'## error ## jsonファイルから読み込んだentityがゼロ件. filename: {filename}, directory: {directory}.')
        if not entities2[0].get('id') or not entities2[0].get('type'):
            err.set_errorobj('13', f'## error ## jsonファイルから読み込んだentityにidまたはtypeが無い. filename: {filename}, directory: {directory}.')
        handles['trace_log'].set_log(0,'02',f'input{str(i+1)}: directory: {directory}, inputfile: {filename}')
        
        for add_ent in entities2:
            if add_ent['type'] != entity_type:
                continue
            matched = False
            for in_ent in entities1:
                if add_ent['id'] == in_ent['id']:
                    in_ent.update(add_ent)
                    matched = True
                    break
            if not matched:
                entities1.append(add_ent)
    
    ## 最後に、jsontypeに合わせてEntityのリストをJSON文に変換して整形します
    if (param['jsontype'] == 'batch'):
        outdic={'actionType': 'append'}
        outdic['entities']=entities1
    elif (param['jsontype']=='array'):
        outdic = entities1
    elif (param['jsontype']=='entity'):
        outdic=entities1[0]
    else:
        handles['error'].set_errorobj( '20', f'invalid jsontype option\n')
    
    handles['trace_log'].set_log(0,'03','outputdir: '+param['outputdir']+', outputfile: '+param['outputfile'])
    # print(json.dumps(outdic,ensure_ascii=False, indent=2))
    
    ## 出来上がったJ辞書をJSONファイルとして出力します
    filename = param['outputfile']
    directory = param['outputdir']
    try:
        CT_json.save(directory, filename, outdic, mode='w')
    except FileNotFoundError as e:
        err.set_errorobj('15', f'## error ## {directory} ディレクトリが見つかりません: {e}')
    except PermissionError as e:
        err.set_errorobj('16', f'{filename} ファイルがオープンできません.\nファイル名を変更するか、このファイルを使っているプログラムを終了してください')
    
    if (run['errorflag']==True):
        handles['trace_log'].set_log(0,'05',f'merge faileded')
        print('Error occurd. Please, check message, parameters and files\n')
        input(run['errorobj'])
    if run['warning'] == True:
        handles['trace_log'].__del__()
        input(f'jsonのマージ中にワーニングが発生しました。ログを確認してください\n')
    else:
        handles['trace_log'].set_log(0,'06',f'JSON merged successfully. Number of output Entities: '+str(len(entities1)))
        handles['trace_log'].__del__()
        print(f'json merged successfully. Number of output Entities: {str(len(entities1))}\n')